import { ChildModel, DefineModel, ModelField } from '../src/decorators';
import { App } from 'koishi';
import { mixinModel, registerModel } from '../src/register';
import MemoryDatabase from '@koishijs/plugin-database-memory';
import exp from 'constants';

declare module 'koishi' {
  // eslint-disable-next-line @typescript-eslint/no-empty-interface
  interface User {
    dress: Dress;
    shirt: Shirt;
  }
}

class WearingPreference {
  @ModelField('string(8)')
  color: string;
  @ModelField('string(12)')
  shape: string;

  format() {
    return `${this.color} ${this.shape}`;
  }
}

class Wearing {
  @ModelField('string(3)')
  size: string;

  getSize() {
    return this.size;
  }

  @ChildModel()
  preference: WearingPreference;
}

class Dress extends Wearing {
  getDressSize() {
    return this.getSize();
  }
}
class Shirt extends Wearing {
  getShirtSize() {
    return this.getSize();
  }
}

describe('Model test', () => {
  let app: App;

  beforeEach(async () => {
    app = new App();
    app.plugin(MemoryDatabase);
    mixinModel(app, 'user', { dress: Dress, shirt: Shirt });
    await app.start();
  });

  it('should register model fields', () => {
    const { user } = app.model.config;
    expect(user.fields['dress.size'].type).toBe('string');
    expect(user.fields['dress.size'].length).toBe(3);
    expect(user.fields['shirt.size'].type).toBe('string');
    expect(user.fields['shirt.size'].length).toBe(3);
    expect(user.fields['dress.preference.color'].type).toBe('string');
    expect(user.fields['dress.preference.color'].length).toBe(8);
    expect(user.fields['shirt.preference.color'].type).toBe('string');
    expect(user.fields['shirt.preference.color'].length).toBe(8);
    expect(user.fields['dress.preference.shape'].type).toBe('string');
    expect(user.fields['dress.preference.shape'].length).toBe(12);
    expect(user.fields['shirt.preference.shape'].type).toBe('string');
    expect(user.fields['shirt.preference.shape'].length).toBe(12);

    expect(user.internal['dress.']).toEqual(Dress.prototype);
    expect(user.internal['shirt.']).toEqual(Shirt.prototype);
    expect(user.internal['dress.preference.']).toEqual(
      WearingPreference.prototype,
    );
    expect(user.internal['shirt.preference.']).toEqual(
      WearingPreference.prototype,
    );
  });

  it('should make class instance', async () => {
    await app.database.upsert('user', [
      {
        id: '111111',
        dress: {
          size: 'S',
          preference: {
            color: 'red',
            shape: 'round',
          },
        },
        shirt: {
          size: 'M',
          preference: {
            color: 'blue',
            shape: 'square',
          },
        },
      },
    ]);
    const [user] = await app.database.get('user', { id: '111111' });
    expect(user.dress).toBeInstanceOf(Dress);
    expect(user.dress.getDressSize()).toBe('S');
    expect(user.dress.preference).toBeInstanceOf(WearingPreference);
    expect(user.dress.preference.format()).toBe('red round');
    expect(user.shirt).toBeInstanceOf(Shirt);
    expect(user.shirt.getShirtSize()).toBe('M');
    expect(user.shirt.preference).toBeInstanceOf(WearingPreference);
    expect(user.shirt.preference.format()).toBe('blue square');
  });
});
